#version 410 core

layout (triangles_adjacency) in;    // six vertices in
layout (triangle_strip, max_vertices = 18) out; // 4 per quad * 3 triangle vertices + 6 for near/far caps

in vec3 PosL[];

uniform vec3 lightPosition;
uniform mat4 gWVP;
uniform mat4 M;
uniform mat4 VP;

float EPSILON = 0.0;

void EmitQuad(vec3 StartVertex, vec3 EndVertex, vec3 transformedLightPosition)
{
    // Vertex 1
    vec3 LightDir = normalize(StartVertex - transformedLightPosition);   
    gl_Position = VP * M * vec4((StartVertex + LightDir * EPSILON), 1.0);
    EmitVertex();
 
    // Vertex 2
    gl_Position = VP * M * vec4(LightDir, 0.0);
    EmitVertex();
    
    // Vertex 3
    LightDir = normalize(EndVertex - transformedLightPosition);
    gl_Position = VP * M * vec4((EndVertex + LightDir * EPSILON), 1.0);
    EmitVertex();
    
    // Vertex 4
    gl_Position = VP * M * vec4(LightDir , 0.0);
    EmitVertex();

    EndPrimitive();            
}


void main()
{
    vec3 e1 = PosL[2] - PosL[0];
    vec3 e2 = PosL[4] - PosL[0];
    vec3 e3 = PosL[1] - PosL[0];
    vec3 e4 = PosL[3] - PosL[2];
    vec3 e5 = PosL[4] - PosL[2];
    vec3 e6 = PosL[5] - PosL[0];

	mat4 inverseModelMatrix = inverse(M);
	vec3 transformedLightPosition = vec3(inverseModelMatrix * vec4(lightPosition, 1.0));
    vec3 Normal = normalize(cross(e1,e2));
	vec3 LightDir = normalize(transformedLightPosition - PosL[0]);

    // only light facing triangles
    if (dot(Normal, LightDir) > 0) {

        Normal = cross(e3,e1);

        if (dot(Normal, LightDir) <= 0) {
            vec3 StartVertex = PosL[0];
            vec3 EndVertex = PosL[2];
            EmitQuad(StartVertex, EndVertex, transformedLightPosition);
        }

        Normal = cross(e4,e5);
        LightDir = transformedLightPosition - PosL[2];

        if (dot(Normal, LightDir) <= 0) {
            vec3 StartVertex = PosL[2];
            vec3 EndVertex = PosL[4];
            EmitQuad(StartVertex, EndVertex, transformedLightPosition);
        }

        Normal = cross(e2,e6);
        LightDir = transformedLightPosition - PosL[4];

        if (dot(Normal, LightDir) <= 0) {
            vec3 StartVertex = PosL[4];
            vec3 EndVertex = PosL[0];
            EmitQuad(StartVertex, EndVertex, transformedLightPosition);
        }

        // render the front cap
        LightDir = (normalize(PosL[0] - transformedLightPosition));
        gl_Position = VP * M * vec4((PosL[0] + LightDir * EPSILON), 1.0);
        EmitVertex();

        LightDir = (normalize(PosL[2] - transformedLightPosition));
        gl_Position = VP * M * vec4((PosL[2] + LightDir * EPSILON), 1.0);
        EmitVertex();

        LightDir = (normalize(PosL[4] - transformedLightPosition));
        gl_Position = VP * M * vec4((PosL[4] + LightDir * EPSILON), 1.0);
        EmitVertex();
        EndPrimitive();
 
        // render the back cap
        LightDir = PosL[0] - transformedLightPosition;
        gl_Position = VP * M * vec4(LightDir, 0.0);
        EmitVertex();

        LightDir = PosL[4] - transformedLightPosition;
        gl_Position = VP * M * vec4(LightDir, 0.0);
        EmitVertex();

        LightDir = PosL[2] - transformedLightPosition;
        gl_Position = VP * M * vec4(LightDir, 0.0);
        EmitVertex();
        EndPrimitive();
    }
}